/*
 * @(#)Charset.java	1.32 02/01/02
 *
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package java.nio.charset;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.spi.CharsetProvider;
import java.security.AccessController;
import java.security.AccessControlException;
import java.security.PrivilegedAction;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import sun.misc.Service;
import sun.misc.ServiceConfigurationError;
import sun.nio.cs.StandardCharsets;
import sun.nio.cs.ThreadLocalCoders;


/**
 * A named mapping between sequences of sixteen-bit Unicode characters and
 * sequences of bytes.  This class defines methods for creating decoders and
 * encoders and for retrieving the various names associated with a charset.
 * Instances of this class are immutable.
 *
 * <p> This class also defines static methods for testing whether a particular
 * charset is supported, for locating charset instances by name, and for
 * constructing a map that contains every charset for which support is
 * available in the current Java virtual machine.  Support for new charsets can
 * be added via the service-provider interface defined in the {@link
 * java.nio.charset.spi.CharsetProvider} class.
 *
 * <p> All of the methods defined in this class are safe for use by multiple
 * concurrent threads.
 *
 *
 * <a name="names"><a name="charenc">
 * <h4>Charset names</h4>
 *
 * <p> Charsets are named by strings composed of the following characters:
 *
 * <ul>
 *
 *   <li> The uppercase letters <tt>'A'</tt> through <tt>'Z'</tt>
 *        (<tt>'&#92;u0041'</tt>&nbsp;through&nbsp;<tt>'&#92;u005a'</tt>),
 * 
 *   <li> The lowercase letters <tt>'a'</tt> through <tt>'z'</tt>
 *        (<tt>'&#92;u0061'</tt>&nbsp;through&nbsp;<tt>'&#92;u007a'</tt>),
 * 
 *   <li> The digits <tt>'0'</tt> through <tt>'9'</tt>
 *        (<tt>'&#92;u0030'</tt>&nbsp;through&nbsp;<tt>'&#92;u0039'</tt>),
 * 
 *   <li> The dash character <tt>'-'</tt>
 *        (<tt>'&#92;u002d'</tt>,&nbsp;<small>HYPHEN-MINUS</small>),
 *
 *   <li> The period character <tt>'.'</tt>
 *        (<tt>'&#92;u002e'</tt>,&nbsp;<small>FULL STOP</small>),
 * 
 *   <li> The colon character <tt>':'</tt>
 *        (<tt>'&#92;u003a'</tt>,&nbsp;<small>COLON</small>), and
 * 
 *   <li> The underscore character <tt>'_'</tt>
 *        (<tt>'&#92;u005f'</tt>,&nbsp;<small>LOW&nbsp;LINE</small>).
 * 
 * </ul>
 *
 * A charset name must begin with either a letter or a digit.  The empty string
 * is not a legal charset name.  Charset names are not case-sensitive; that is,
 * case is always ignored when comparing charset names.  Charset names
 * generally follow the conventions documented in <a
 * href="http://ietf.org/rfc/rfc2278.txt"><i>RFC&nbsp;2278:&nbsp;IANA Charset
 * Registration Procedures</i></a>.
 *
 * <p> Every charset has a <i>canonical name</i> and may also have one or more
 * <i>aliases</i>.  The canonical name is returned by {@link #name name} method
 * of this class.  Canonical names are, by convention, usually in upper case.
 * The aliases of a charset are returned by the {@link #aliases aliases}
 * method.
 *
 * <a name="hn">
 *
 * <p> Some charsets have an <i>historical name</i> that is defined for
 * compatibility with previous versions of the Java platform.  A charset's
 * historical name is either its canonical name or one of its aliases.  The
 * historical name is returned by the <tt>getEncoding()</tt> methods of the
 * {@link java.io.InputStreamReader#getEncoding InputStreamReader} and {@link
 * java.io.OutputStreamWriter#getEncoding OutputStreamWriter} classes.
 *
 * <a name="iana">
 *
 * <p> If a charset listed in the <a
 * href="http://www.iana.org/assignments/character-sets"><i>IANA Charset
 * Registry</i></a> is supported by an implementation of the Java platform then
 * its canonical name must be the name listed in the registry.  Many charsets
 * are given more than one name in the registry, in which case the registry
 * identifies one of the names as <i>MIME-preferred</i>.  If a charset has more
 * than one registry name then its canonical name must be the MIME-preferred
 * name and the other names in the registry must be valid aliases.  If a
 * supported charset is not listed in the IANA registry then its canonical name
 * must begin with one of the strings <tt>"X-"</tt> or <tt>"x-"</tt>.
 *
 * <p> The IANA charset registry does change over time, and so the canonical
 * name and the aliases of a particular charset may also change over time.  To
 * ensure compatibility it is recommended that no alias ever be removed from a
 * charset, and that if the canonical name of a charset is changed then its
 * previous canonical name be made into an alias.
 *
 *
 * <h4>Standard charsets</h4>
 *
 * <p> Every implementation of the Java platform is required to support the
 * following standard charsets.  Consult the release documentation for your
 * implementation to see if any other charsets are supported.
 * 
 * <blockquote><table width="80%">
 * <tr><td valign=top><tt>US-ASCII</tt></td>
 *     <td>Seven-bit ASCII, a.k.a. <tt>ISO646-US</tt>,
 *         a.k.a. the Basic Latin block of the Unicode character set</td></tr>
 * <tr><td valign=top><tt>ISO-8859-1&nbsp;&nbsp;</tt></td>
 *     <td>ISO Latin Alphabet No. 1, a.k.a. <tt>ISO-LATIN-1</tt></td></tr>
 * <tr><td valign=top><tt>UTF-8</tt></td>
 *     <td>Eight-bit UCS Transformation Format</td></tr>
 * <tr><td valign=top><tt>UTF-16BE</tt></td>
 *     <td>Sixteen-bit UCS Transformation Format,
 *         big-endian byte&nbsp;order</td></tr>
 * <tr><td valign=top><tt>UTF-16LE</tt></td>
 *     <td>Sixteen-bit UCS Transformation Format,
 *         little-endian byte&nbsp;order</td></tr>
 * <tr><td valign=top><tt>UTF-16</tt></td>
 *     <td>Sixteen-bit UCS Transformation Format,
 *         byte&nbsp;order identified by an optional byte-order mark</td></tr>
 * </table></blockquote>
 * 
 * <p> The <tt>UTF-8</tt> charset is specified by <a
 * href="http://ietf.org/rfc/rfc2279.txt"><i>RFC&nbsp;2279</i></a>; the
 * transformation format upon which it is based is specified in
 * Amendment&nbsp;2 of ISO&nbsp;10646-1 and is also described in
 * &#167;&nbsp;3.8 of <a
 * href="http://www.unicode.org/unicode/standard/standard.html"><i>The Unicode
 * Standard, Version&nbsp;3.0</i></a> (<a
 * href="http://www.unicode.org/unicode/uni2errata/UTF-8_Corrigendum.html">amended</a>).
 *
 * <p> The <tt>UTF-16</tt> charsets are specified by <a
 * href="http://ietf.org/rfc/rfc2781.txt"><i>RFC&nbsp;2781</i></a>; the
 * transformation formats upon which they are based are specified in
 * Amendment&nbsp;1 of ISO&nbsp;10646-1 and are also described in
 * &#167;&nbsp;3.8 of <a
 * href="http://www.unicode.org/unicode/standard/standard.html"><i>The Unicode
 * Standard, Version&nbsp;3.0</i></a>.
 *
 * <p> The <tt>UTF-16</tt> charsets use sixteen-bit quantities and are
 * therefore sensitive to byte order.  In these encodings the byte order of a
 * stream may be indicated by an initial <i>byte-order mark</i> represented by
 * the Unicode character <tt>'&#92;uFEFF'</tt>.  Byte-order marks are handled
 * as follows:
 *
 * <ul>
 *
 *   <li><p> When decoding, the <tt>UTF-16BE</tt> and <tt>UTF-16LE</tt>
 *   charsets ignore byte-order marks; when encoding, they do not write
 *   byte-order marks. </p></li>
 *
 *   <li><p> When decoding, the <tt>UTF-16</tt> charset interprets a byte-order
 *   mark to indicate the byte order of the stream but defaults to big-endian
 *   if there is no byte-order mark; when encoding, it uses big-endian byte
 *   order and writes a big-endian byte-order mark. </p></li>
 *
 * </ul>
 *
 * In any case, when a byte-order mark is read at the beginning of a decoding
 * operation it is omitted from the resulting sequence of characters.  Byte
 * order marks occuring after the first element of an input sequence are not
 * omitted since the same code is used to represent <small>ZERO-WIDTH
 * NON-BREAKING SPACE</small>.
 *
 * <p> Every instance of the Java virtual machine has a default charset, which
 * may or may not be one of the standard charsets.  The default charset is
 * determined during virtual-machine startup and typically depends upon the
 * locale and charset being used by the underlying operating system. </p>
 *
 *
 * <h4>Terminology</h4>
 *
 * <p> The name of this class is taken from the terms used in <a
 * href="http://ietf.org/rfc/rfc2278.txt""><i>RFC&nbsp;2278</i></a>.  In that
 * document a <i>charset</i> is defined as the combination of a coded character
 * set and a character-encoding scheme.
 *
 * <p> A <i>coded character set</i> is a mapping between a set of abstract
 * characters and a set of integers.  US-ASCII, ISO&nbsp;8859-1,
 * JIS&nbsp;X&nbsp;0201, and full Unicode, which is the same as
 * ISO&nbsp;10646-1, are examples of coded character sets.
 *
 * <p> A <i>character-encoding scheme</i> is a mapping between a coded
 * character set and a set of octet (eight-bit byte) sequences.  UTF-8, UCS-2,
 * UTF-16, ISO&nbsp;2022, and EUC are examples of character-encoding schemes.
 * Encoding schemes are often associated with a particular coded character set;
 * UTF-8, for example, is used only to encode Unicode.  Some schemes, however,
 * are associated with multiple character sets; EUC, for example, can be used
 * to encode characters in a variety of Asian character sets.
 *
 * <p> When a coded character set is used exclusively with a single
 * character-encoding scheme then the corresponding charset is usually named
 * for the character set; otherwise a charset is usually named for the encoding
 * scheme and, possibly, the locale of the character sets that it supports.
 * Hence <tt>US-ASCII</tt> is the name of the charset for US-ASCII while
 * <tt>EUC-JP</tt> is the name of the charset that encodes the
 * JIS&nbsp;X&nbsp;0201, JIS&nbsp;X&nbsp;0208, and JIS&nbsp;X&nbsp;0212
 * character sets.
 *
 * <p> The native coded character set of the Java programming language is that
 * of the first seventeen planes of the Unicode version&nbsp;3.0 character set;
 * that is, it consists in the <i>basic multilingual plane</i> (BMP) of Unicode
 * version&nbsp;1 plus the next sixteen planes of Unicode version&nbsp;3.  This
 * is because the language's internal representation of characters uses the
 * UTF-16 encoding, which encodes the BMP directly and uses <i>surrogate
 * pairs</i>, a simple escape mechanism, to encode the other planes.  Hence a
 * charset in the Java platform defines a mapping between sequences of
 * sixteen-bit values in UTF-16 and sequences of bytes. </p>
 *
 *
 * @author Mark Reinhold
 * @author JSR-51 Expert Group
 * @version 1.32, 02/01/02
 * @since 1.4
 *
 * @see CharsetDecoder
 * @see CharsetEncoder
 * @see java.nio.charset.spi.CharsetProvider
 */

public abstract class Charset
    implements Comparable
{

    /* -- Static methods -- */

    /**
     * Checks that the given string is a legal charset name. </p>
     *
     * @param  s
     *         A purported charset name
     *
     * @throws  IllegalCharsetNameException
     *          If the given name is not a legal charset name
     */
    private static void checkName(String s) {
	int n = s.length();
	for (int i = 0; i < n; i++) {
	    char c = s.charAt(i);
	    if (c >= 'A' && c <= 'Z') continue;
	    if (c >= 'a' && c <= 'z') continue;
	    if (c >= '0' && c <= '9') continue;
	    if (c == '-') continue;
	    if (c == ':') continue;
	    if (c == '_') continue;
	    if (c == '.') continue;
	    throw new IllegalCharsetNameException(s);
	}
    }

    /* The standard set of charsets */
    private static CharsetProvider standardProvider = new StandardCharsets();

    // Cache of the most-recently-returned charset,
    // along with the name that was used to find it
    //
    private static volatile Object[] cache = null;

    private static Charset cache(String charsetName, Charset cs) {
	cache = new Object[] { charsetName, cs };
	return cs;
    }

    // Creates an iterator that walks over the available providers, ignoring
    // those whose lookup or instantiation causes a security exception to be
    // thrown.  Should be invoked with full privileges.
    //
    private static Iterator providers() {
	return new Iterator() {

		Class c = java.nio.charset.spi.CharsetProvider.class;
		ClassLoader cl = ClassLoader.getSystemClassLoader();
		Iterator i = Service.providers(c, cl);
		Object next = null;

		private boolean getNext() {
		    while (next == null) {
			try {
			    if (!i.hasNext())
				return false;
			    next = i.next();
			} catch (ServiceConfigurationError sce) {
			    if (sce.getCause() instanceof SecurityException) {
				// Ignore security exceptions
				continue;
			    }
			    throw sce;
			}
		    }
		    return true;
		}

		public boolean hasNext() {
		    return getNext();
		}

		public Object next() {
		    if (!getNext())
			throw new NoSuchElementException();
		    Object n = next;
		    next = null;
		    return n;
		}

		public void remove() {
		    throw new UnsupportedOperationException();
		}

	    };
    }

    // Thread-local gate to prevent recursive provider lookups
    private static ThreadLocal gate = new ThreadLocal();

    private static Charset lookupViaProviders(final String charsetName) {

	// The runtime startup sequence looks up standard charsets as a
	// consequence of the VM's invocation of System.initializeSystemClass
	// in order to, e.g., set system properties and encode filenames.  At
	// that point the application class loader has not been initialized,
	// however, so we can't look for providers because doing so will cause
	// that loader to be prematurely initialized with incomplete
	// information.
	//
	if (!sun.misc.VM.isBooted())
	    return null;

	if (gate.get() != null)
	    // Avoid recursive provider lookups
	    return null;
	try {
	    gate.set(gate);

	    return (Charset)AccessController
		.doPrivileged(new PrivilegedAction() {
		    public Object run() {
			for (Iterator i = providers(); i.hasNext();) {
			    CharsetProvider cp = (CharsetProvider)i.next();
			    Charset cs = cp.charsetForName(charsetName);
			    if (cs != null)
				return cs;
			}
			return null;
		    }
		});

	} finally {
	    gate.set(null);
	}
    }

    private static Charset lookup(String charsetName) {
	if (charsetName == null)
	    throw new IllegalArgumentException("Null charset name");
	Object[] ca = cache;
	if ((ca != null) && ca[0].equals(charsetName))
	    return (Charset)ca[1];
	Charset cs = standardProvider.charsetForName(charsetName);
	if (cs != null)
	    return cache(charsetName, cs);
	cs = lookupViaProviders(charsetName);
	if (cs != null)
	    return cache(charsetName, cs);
	/* Only need to check the name if we didn't find a charset for it */
	checkName(charsetName);
	return null;
    }

    /**
     * Tells whether the named charset is supported. </p>
     *
     * @param  charsetName
     *         The name of the requested charset; may be either
     *         a canonical name or an alias
     *
     * @return  <tt>true</tt> if, and only if, support for the named charset
     *          is available in the current Java virtual machine
     *
     * @throws IllegalCharsetNameException
     *         If the given charset name is illegal
     */
    public static boolean isSupported(String charsetName) {
	return (lookup(charsetName) != null);
    }

    /**
     * Returns a charset object for the named charset. </p>
     *
     * @param  charsetName
     *         The name of the requested charset; may be either
     *         a canonical name or an alias
     *
     * @return  A charset object for the named charset
     *
     * @throws  IllegalCharsetNameException
     *          If the given charset name is illegal
     *
     * @throws  UnsupportedCharsetException
     *          If no support for the named charset is available
     *          in this instance of the Java virtual machine
     */
    public static Charset forName(String charsetName) {
	Charset cs = lookup(charsetName);
	if (cs != null)
	    return cs;
	throw new UnsupportedCharsetException(charsetName);
    }

    // Fold charsets from the given iterator into the given map, ignoring
    // charsets whose names already have entries in the map.
    //
    private static void put(Iterator i, Map m) {
	while (i.hasNext()) {
	    Charset cs = (Charset)i.next();
	    if (!m.containsKey(cs.name()))
		m.put(cs.name(), cs);
	}
    }

    /**
     * Constructs a sorted map from canonical charset names to charset objects.
     *
     * <p> The map returned by this method will have one entry for each charset
     * for which support is available in the current Java virtual machine.  If
     * two or more supported charsets have the same canonical name then the
     * resulting map will contain just one of them; which one it will contain
     * is not specified. </p>
     *
     * <p> The invocation of this method, and the subsequent use of the
     * resulting map, may cause time-consuming disk or network I/O operations
     * to occur.  This method is provided for applications that need to
     * enumerate all of the available charsets, for example to allow user
     * charset selection.  This method is not used by the {@link #forName
     * forName} method, which instead employs an efficient incremental lookup
     * algorithm.
     *
     * <p> This method may return different results at different times if new
     * charset providers are dynamically made available to the current Java
     * virtual machine.  In the absence of such changes, the charsets returned
     * by this method are exactly those that can be retrieved via the {@link
     * #forName forName} method.  </p>
     *
     * @return An immutable, case-insensitive map from canonical charset names
     *         to charset objects
     */
    public static SortedMap availableCharsets() {
	return (SortedMap)AccessController
	    .doPrivileged(new PrivilegedAction() {
		public Object run() {
		    TreeMap m = new TreeMap(String.CASE_INSENSITIVE_ORDER);
		    put(standardProvider.charsets(), m);
		    for (Iterator i = providers(); i.hasNext();) {
			CharsetProvider cp = (CharsetProvider)i.next();
			put(cp.charsets(), m);
		    }
		    return Collections.unmodifiableSortedMap(m);
		}
	    });
    }


    /* -- Instance fields and methods -- */

    private final String name;		// tickles a bug in oldjavac
    private final String[] aliases;	// tickles a bug in oldjavac
    private Set aliasSet = null;

    /**
     * Initializes a new charset with the given canonical name and alias
     * set. </p>
     *
     * @param  canonicalName
     *         The canonical name of this charset
     *
     * @param  aliases
     *         An array of this charset's aliases, or null if it has no aliases
     *
     * @throws IllegalCharsetNameException
     *         If the canonical name or any of the aliases are illegal
     */
    protected Charset(String canonicalName, String[] aliases) {
	checkName(canonicalName);
	String[] as = (aliases == null) ? new String[0] : aliases;
	for (int i = 0; i < as.length; i++)
	    checkName(as[i]);
	this.name = canonicalName;
	this.aliases = as;
    }

    /**
     * Returns this charset's canonical name. </p>
     *
     * @return  The canonical name of this charset
     */
    public final String name() {
	return name;
    }

    /**
     * Returns a set containing this charset's aliases. </p>
     *
     * @return  An immutable set of this charset's aliases
     */
    public final Set aliases() {
	if (aliasSet != null)
	    return aliasSet;
	int n = aliases.length;
	HashSet hs = new HashSet(n);
	for (int i = 0; i < n; i++)
	    hs.add(aliases[i]);
	aliasSet = Collections.unmodifiableSet(hs);
	return aliasSet;
    }

    /**
     * Returns this charset's human-readable name for the default locale.
     *
     * <p> The default implementation of this method simply returns this
     * charset's canonical name.  Concrete subclasses of this class may
     * override this method in order to provide a localized display name. </p>
     *
     * @return  The display name of this charset in the default locale
     */
    public String displayName() {
	return name;
    }

    /**
     * Tells whether or not this charset is registered in the <a
     * href="http://www.iana.org/assignments/character-sets">IANA Charset
     * Registry</a>.  </p>
     *
     * @return  <tt>true</tt> if, and only if, this charset is known by its
     *          implementor to be registered with the IANA
     */
    public final boolean isRegistered() {
	return !name.startsWith("X-") && !name.startsWith("x-");
    }

    /**
     * Returns this charset's human-readable name for the given locale.
     *
     * <p> The default implementation of this method simply returns this
     * charset's canonical name.  Concrete subclasses of this class may
     * override this method in order to provide a localized display name. </p>
     *
     * @param  locale
     *         The locale for which the display name is to be retrieved
     *
     * @return  The display name of this charset in the given locale
     */
    public String displayName(Locale locale) {
	return name;
    }

    /**
     * Tells whether or not this charset contains the given charset.
     *
     * <p> A charset <i>C</i> is said to <i>contain</i> a charset <i>D</i> if,
     * and only if, every character representable in <i>D</i> is also
     * representable in <i>C</i>.  If this relationship holds then it is
     * guaranteed that every string that can be encoded in <i>D</i> can also be
     * encoded in <i>C</i> without performing any replacements.
     *
     * <p> That <i>C</i> contains <i>D</i> does not imply that each character
     * representable in <i>C</i> by a particular byte sequence is represented
     * in <i>D</i> by the same byte sequence, although sometimes this is the
     * case.
     *
     * <p> Every charset contains itself.
     *
     * <p> This method computes an approximation of the containment relation:
     * If it returns <tt>true</tt> then the given charset is known to be
     * contained by this charset; if it returns <tt>false</tt>, however, then
     * it is not necessarily the case that the given charset is not contained
     * in this charset.
     *
     * @return  <tt>true</tt> if, and only if, the given charset
     *          is contained in this charset
     */
    public abstract boolean contains(Charset cs);

    /**
     * Constructs a new decoder for this charset. </p>
     *
     * @return  A new decoder for this charset
     */
    public abstract CharsetDecoder newDecoder();

    /**
     * Constructs a new encoder for this charset. </p>
     *
     * @return  A new encoder for this charset
     *
     * @throws  UnsupportedOperationException
     *          If this charset does not support encoding
     */
    public abstract CharsetEncoder newEncoder();

    /**
     * Tells whether or not this charset supports encoding.
     *
     * <p> Nearly all charsets support encoding.  The primary exceptions are
     * special-purpose <i>auto-detect</i> charsets whose decoders can determine
     * which of several possible encoding schemes is in use by examining the
     * input byte sequence.  Such charsets do not support encoding because
     * there is no way to determine which encoding should be used on output.
     * Implementations of such charsets should override this method to return
     * <tt>false</tt>. </p>
     *
     * @return  <tt>true</tt> if, and only if, this charset supports encoding
     */
    public boolean canEncode() {
	return true;
    }

    /**
     * Convenience method that decodes bytes in this charset into Unicode
     * characters.
     *
     * <p> An invocation of this method upon a charset <tt>cs</tt> returns the
     * same result as the expression
     *
     * <pre>
     *     cs.newDecoder()
     *       .onMalformedInput(CodingErrorAction.REPLACE)
     *       .onUnmappableCharacter(CodingErrorAction.REPLACE)
     *       .decode(bb); </pre>
     *
     * except that it is potentially more efficient because it can cache
     * decoders between successive invocations.
     *
     * <p> This method always replaces malformed-input and unmappable-character
     * sequences with this charset's default replacement byte array.  In order
     * to detect such sequences, use the {@link
     * CharsetDecoder#decode(java.nio.ByteBuffer)} method directly.  </p>
     *
     * @param  bb  The byte buffer to be decoded
     *
     * @return  A char buffer containing the decoded characters
     */
    public final CharBuffer decode(ByteBuffer bb) {
	try {
	    return ThreadLocalCoders.decoderFor(this)
		.onMalformedInput(CodingErrorAction.REPLACE)
		.onUnmappableCharacter(CodingErrorAction.REPLACE)
		.decode(bb);
	} catch (CharacterCodingException x) {
	    throw new Error(x);		// Can't happen
	}
    }

    /**
     * Convenience method that encodes Unicode characters into bytes in this
     * charset.
     *
     * <p> An invocation of this method upon a charset <tt>cs</tt> returns the
     * same result as the expression
     *
     * <pre>
     *     cs.newEncoder()
     *       .onMalformedInput(CodingErrorAction.REPLACE)
     *       .onUnmappableCharacter(CodingErrorAction.REPLACE)
     *       .encode(bb); </pre>
     *
     * except that it is potentially more efficient because it can cache
     * encoders between successive invocations.
     *
     * <p> This method always replaces malformed-input and unmappable-character
     * sequences with this charset's default replacement string.  In order to
     * detect such sequences, use the {@link
     * CharsetEncoder#encode(java.nio.CharBuffer)} method directly.  </p>
     *
     * @param  bb  The char buffer to be encoded
     *
     * @return  A byte buffer containing the encoded characters
     */
    public final ByteBuffer encode(CharBuffer cb) {
	try {
	    return ThreadLocalCoders.encoderFor(this)
		.onMalformedInput(CodingErrorAction.REPLACE)
		.onUnmappableCharacter(CodingErrorAction.REPLACE)
		.encode(cb);
	} catch (CharacterCodingException x) {
	    throw new Error(x);		// Can't happen
	}
    }

    /**
     * Convenience method that encodes a string into bytes in this charset.
     *
     * <p> An invocation of this method upon a charset <tt>cs</tt> returns the
     * same result as the expression
     *
     * <pre>
     *     cs.encode(CharBuffer.wrap(s)); </pre>
     *
     * @param  str  The string to be encoded
     *
     * @return  A byte buffer containing the encoded characters
     */
    public final ByteBuffer encode(String str) {
	return encode(CharBuffer.wrap(str));
    }

    /**
     * Compares this charset to another object.
     *
     * <p> Charsets are ordered by their canonical names, without regard to
     * case. </p>
     *
     * @param  ob
     *         The object to which this object is to be compared
     *
     * @return A negative integer, zero, or a positive integer as this charset
     *         is less than, equal to, or greater than the specified object
     */
    public final int compareTo(Object ob) {
	return (name().compareToIgnoreCase(((Charset)ob).name()));
    }

    /**
     * Computes a hashcode for this charset. </p>
     *
     * @return  An integer hashcode
     */
    public final int hashCode() {
	return name().hashCode();
    }

    /**
     * Tells whether or not this object is equal to another.
     *
     * <p> Two charsets are equal if, and only if, they have the same canonical
     * names.  A charset is never equal to any other type of object.  </p>
     *
     * @return  <tt>true</tt> if, and only if, this charset is equal to the
     *          given object
     */
    public final boolean equals(Object ob) {
	if (!(ob instanceof Charset))
	    return false;
	if (this == ob)
	    return true;
	return name.equals(((Charset)ob).name());
    }

    /**
     * Returns a string describing this charset. </p>
     *
     * @return  A string describing this charset
     */
    public final String toString() {
	return name();
    }

}
